package org.ten60.netkernel.security.endpoint;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.net.URI;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.util.Comparator;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.ten60.netkernel.security.util.SecurityUtils;

public class SignVerifyModuleAccessor extends StandardAccessorImpl
{	
	public final static String MODULE_SIGNATURE_FILE="module.signature";
	
	public SignVerifyModuleAccessor()
	{	this.declareThreadSafe();
	}

	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	String type=context.getThisRequest().getArgumentValue("activeType");
		String moduleURI=context.getThisRequest().getArgumentValue("operand");
		Object result=null;
		if(type.equals("pkiSignModule"))
		{	result=signModule(context, moduleURI);
		}
		else
		{	result=verifyModule(context, moduleURI);
		}
		context.createResponseFrom(result);
	}
	
	private Object signModule(INKFRequestContext context, String moduleURI) throws Exception
	{	DigestValue digest=createDigest(context, moduleURI);
		INKFRequest req=createPartialRequest(context, "active:pkiSign", digest);
		req.addArgument("keyPassword", context.getThisRequest().getArgumentValue("keyPassword"));
		return context.issueRequest(req);
	}

	private Object verifyModule(INKFRequestContext context, String moduleURI) throws Exception
	{	DigestValue digest=createDigest(context, moduleURI);
		INKFRequest req=createPartialRequest(context, "active:pkiVerify", digest);
		if(context.getThisRequest().argumentExists("signature"))
		{	req.addArgument("signature", context.getThisRequest().getArgumentValue("signature"));
		}
		else
		{	if(digest.mSignature!=null)
			{	req.addArgumentByValue("signature", digest.mSignature);
			}
			else
			{	req.addArgument("signature", moduleURI+MODULE_SIGNATURE_FILE);
			}
		}
		return context.issueRequest(req);	
	}
	
	private INKFRequest createPartialRequest(INKFRequestContext context, String base, DigestValue digest) throws Exception
	{	INKFRequest req=context.createRequest(base);
		req.addArgumentByValue("operand", new ByteArrayRepresentation(digest.mDigest));
		if(context.getThisRequest().argumentExists("publicKeyCertificate"))
		{	req.addArgument("publicKeyCertificate", context.getThisRequest().getArgumentValue("publicKeyCertificate"));
		}
		else
		{	req.addArgument("keystore", context.getThisRequest().getArgumentValue("keystore"));
			req.addArgument("keystorePassword", context.getThisRequest().getArgumentValue("keystorePassword"));
			req.addArgument("keyID", context.getThisRequest().getArgumentValue("keyID"));
		}
		return req;
	}
	
	
	private DigestValue createDigest(INKFRequestContext context, String moduleURI) throws Exception
	{	if(moduleURI.endsWith("jar"))
		{	return createJarDigest(context, moduleURI);			
		}
		URI u=URI.create(moduleURI);
		if(u.getScheme().equals("file"))
		{	return createFileDigest(new File(u));			
		}
		else throw new NKFException("Unsupported module URI  "+moduleURI);
	}
	
	private DigestValue createFileDigest(File f) throws Exception
	{	HDSBuilder b=new HDSBuilder();
		recurseFileTree(f, b);
	
		MessageDigest md = MessageDigest.getInstance("MD5");
		ByteArrayOutputStream baos=new ByteArrayOutputStream(1024);
		DigestOutputStream dos=new DigestOutputStream(baos, md);
	
		IHDSNode hds=b.getRoot();
		IHDSNodeList nodelist=hds.getNodes("/file");
		
		//Sort them to make platform independent...
		nodelist=nodelist.sort(new NodeFileComparator());
		
		for(int i=0; i<nodelist.size(); i++)
		{	f=(File)nodelist.get(i).getValue();
			FileInputStream fis=new FileInputStream(f);
			Utils.pipe(fis, dos);
		}
		dos.flush();
		DigestValue dv=new DigestValue();
		dv.mDigest=md.digest();
		return dv;			
	}
	
	private DigestValue createJarDigest(INKFRequestContext context, String moduleURI) throws Exception
	{	//Create Temporary Directory Structure
		String guid=SecurityUtils.GUID();
		File tmp=File.createTempFile("nk-module-"+guid,".tmp");
		tmp.deleteOnExit();
		File tmpdir=new File(tmp.getParent(), "nk-module"+guid);
		tmpdir.mkdir();
		tmpdir.deleteOnExit();
		tmp.delete();
		tmp=new File(tmpdir, "tmp-nk-module.zip");
		tmp.deleteOnExit();
		//Stream and Unpack to Temp Dir
		IReadableBinaryStreamRepresentation rbs=(IReadableBinaryStreamRepresentation)context.source(moduleURI, IReadableBinaryStreamRepresentation.class);
		Utils.pipe(rbs.getInputStream(), new FileOutputStream(tmp));
		unzip(new FileInputStream(tmp), tmp.toURI().toString());
		tmp.delete();
		DigestValue dv=createFileDigest(tmpdir);
		//Locate signature file
		File signature=new File(tmpdir, MODULE_SIGNATURE_FILE);
		if(signature.exists())
		{	dv.mSignature=(String)context.source(signature.toURI().toString(), String.class);			
		}
		rmdir(tmpdir);
		return dv;
	}

	private void recurseFileTree(File dir, HDSBuilder b)
	{	File[] files=dir.listFiles(new SigFilter());
		if(files!=null)
		{	for(int i=0; i<files.length; i++)
			{	if(files[i].isDirectory())
				{	recurseFileTree(files[i],b);
				}
				else
				{	b.addNode("file",files[i]);
				}
			}
		}
	}
	
	private class NodeFileComparator implements Comparator<IHDSNode>
	{
		public int compare(IHDSNode o1, IHDSNode o2)
		{
			File f1=(File)o1.getValue();
			File f2=(File)o2.getValue();
			return f1.getAbsolutePath().compareTo(f2.getAbsolutePath());
		}			
	}
	
	private class SigFilter implements FileFilter
	{	public boolean accept(File pathname)
		{	return !pathname.getName().equals(MODULE_SIGNATURE_FILE);		
		}
	}
	
	private void rmdir(File f)
	{	if(f.isDirectory())
		{	File[] fs=f.listFiles();
			for(int i=0; i< fs.length; i++)
			{	rmdir(fs[i]);				
			}
		}
		f.delete();
	}
	
	public void unzip(InputStream in, String base) throws Exception
	{	ZipInputStream zin = new ZipInputStream(in);
		ZipEntry e;
		URI b=URI.create(base);
		while((e=zin.getNextEntry())!= null)
		{	if(!e.isDirectory())
			{	File f=new File(b.resolve(e.getName()));
				//System.out.println(f.getAbsolutePath());
				File p=new File(f.getParent());
				p.mkdirs();
				f.createNewFile();
				unzip(zin, f);
			}
		}
		zin.close();
	}
    
	private void unzip(ZipInputStream zin, File f) throws Exception
	{	FileOutputStream out = new FileOutputStream(f);
		byte [] b = new byte[512];
		int len = 0;
		while ( (len=zin.read(b))!= -1 )
		{	out.write(b,0,len);
		}
		out.close();
	}

	private class DigestValue
	{	protected byte[] mDigest;
		protected String mSignature;
	}
	
}
